@@ -1,4 +1,4 @@
-# Copyright (C) 2020 Arm Limited or its affiliates. All rights reserved.
+# Copyright (C) 2020-2021 Arm Limited or its affiliates. All rights reserved.
 #
 # SPDX-License-Identifier: Apache-2.0
 #
@@ -21,8 +21,10 @@
 # Subgraph - Holds a neural network subgraph, pointing at Tensors, Operations, Passes, and CascadedPasses.
 # Graph - A full neural network graph with one or more Subgraphs.
 import enum
+from typing import List
 
 from .operation import Op
+from .shape4d import Shape4D
 
 
 class PassPlacement(enum.Enum):
@@ -36,11 +38,17 @@ class PassPlacement(enum.Enum):
 class TensorAllocator(enum.Enum):
     LinearAlloc = 1
     Greedy = 2
+    HillClimb = 3
 
     def __str__(self):
         return self.name
 
 
+class NetworkType(enum.Enum):
+    TFLite = 1
+    TOSA = 2
+
+
 class Pass:
     def __init__(self, name, placement, is_element_wise, npu_block_type):
         self.inputs = []
@@ -57,6 +65,8 @@ class Pass:
         self.name = name
         self.cascade = None
         self.placement = placement
+        self.ifm_shapes: List[Shape4D] = []
+        self.ofm_shapes: List[Shape4D] = []
 
         # TODO: rename is_element_wise because it is not the same as an ElementWise operator. It is used by the tensor
         # allocation and requires that the OFM and IFM has the exact same address. Essentially complete overlap.
@@ -64,6 +74,7 @@ class Pass:
         self.npu_block_type = npu_block_type
         self.block_config = None  # will be filled in by scheduler
         self.shared_buffer = None  # will be filled in by scheduler
+        self.scheduling_info = None  # will be filled in by scheduler
 
         self.predecessors = []
         self.successors = []
@@ -118,6 +129,8 @@ class CascadedPass:
 
         self.predecessors = []
         self.successors = []
+        self.sram_used = 0
+        self.time = 0
 
     def __str__(self):
         return "<nng.CascadedPass strategy=%s x %s '%s',  passes=%s, block_configs=%s>" % (
@@ -144,7 +157,10 @@ class Subgraph:
         self.command_stream_tensor = None
         self.flash_tensor = None
         # Scratch information locally used in the scheduler
-        self.scheduling_info = {}
+        self.schedule = None
+        self.sched_ops = []
+
+        self.generated_stream_id = None
 
         self.memory_used = {}
         self.memory_used_per_type = {}
@@ -178,7 +194,7 @@ class Subgraph:
                 visit_tensor(inp)
                 inp.consumer_list.append(op)
 
-            if op.type in set((Op.Placeholder, Op.SubgraphInput)):
+            if op.type in (Op.Placeholder, Op.SubgraphInput):
                 assert len(op.outputs) == 1
                 self.input_tensors.append(op.outputs[0])
 
@@ -321,7 +337,7 @@ class Subgraph:
         all_ops = self.get_all_ops()
         unique_ops = []
         for op in all_ops:
-            if op.type in set((Op.Const, Op.Identity, Op.Placeholder)):
+            if op.type in (Op.Const, Op.Identity, Op.Placeholder):
                 continue
 
             attrs = op.attrs.copy()
@@ -330,10 +346,13 @@ class Subgraph:
                 attrs["kshape"] = [kshape[0], kshape[1]]
             attrs["type"] = op.type.name
             attrs.pop("use_cudnn_on_gpu", None)
+            custom_options = attrs.pop("custom_options", None)
             if attrs not in unique_ops:
                 unique_ops.append(attrs)
                 # print attributes in human readable format
                 a = attrs.copy()
+                if custom_options is not None:
+                    a["custom_options"] = custom_options
                 s = a.pop("type")
                 data_format = a.pop("data_format", None)
                 if data_format and data_format != b"NHWC":
@@ -349,7 +368,9 @@ class Subgraph:
                     s += "->" + str(dstt)[9:-2]
                 print(s + " " + str(a))
 
-    def print_graph(self):
+    def print_graph(self, label=None):
+        if label:
+            print(f"\n[ {label} ]")
         print("print_graph()", self.name)
         all_ops = self.get_all_ops()
         for idx, op in enumerate(all_ops):
@@ -361,14 +382,17 @@ class Subgraph:
         for idx, op in enumerate(all_ops):
             print(idx, op.type, op.name)
             for idx, tens in enumerate(op.inputs):
-                print(
-                    "    Input  %02d %20s %20s %20s %s"
-                    % (idx, tens.purpose.name, tens.mem_area.name, tens.mem_type.name, tens)
-                )
+                if tens:
+                    print(
+                        f"    Input  {idx:02d}"
+                        f" {tens.purpose.name:>20} {tens.mem_area.name:>20} {tens.mem_type.name:>20} {tens}"
+                    )
+                else:
+                    print(f"    Input  {idx:02d} {'-':>20} {'-':>20} {'-':>20} {tens}")
             for idx, tens in enumerate(op.outputs):
                 print(
-                    "    Output %02d %20s %20s %20s %s"
-                    % (idx, tens.purpose.name, tens.mem_area.name, tens.mem_type.name, tens)
+                    f"    Output {idx:02d}"
+                    f" {tens.purpose.name:>20} {tens.mem_area.name:>20} {tens.mem_type.name:>20} {tens}"
                 )
             print()
 
@@ -378,22 +402,25 @@ class Subgraph:
         for idx, op in enumerate(all_ops):
             print(idx, op.type, op.name)
             for idx, tens in enumerate(op.inputs):
-                q = tens.quantization
-                if q is None:
-                    print("    Input  %02d %10s NO QUANTIZATION INFO %s" % (idx, tens.dtype, tens.name))
+                if tens:
+                    q = tens.quantization
+                    if q is None:
+                        print(f"    Input  {idx:02d} {tens.dtype!s:>10} NO QUANTIZATION INFO {tens.name}")
+                    else:
+                        print(
+                            f"    Input  {idx:02d} {tens.dtype!s:>10}"
+                            f" min={q.min} max={q.max} scale={q.scale_f32!s} zero_point={q.zero_point} {tens.name}"
+                        )
                 else:
-                    print(
-                        "    Input  %02d %10s min=%s max=%s scale=%s zero_point=%s %s"
-                        % (idx, tens.dtype, q.min, q.max, q.scale_f32, q.zero_point, tens.name)
-                    )
+                    print(f"    Input  {idx:02d} {'-':>10} {tens}")
             for idx, tens in enumerate(op.outputs):
                 q = tens.quantization
                 if q is None:
-                    print("    Output %02d %10s NO QUANTIZATION INFO %s" % (idx, tens.dtype, tens.name))
+                    print(f"    Output {idx:02d} {tens.dtype!s:>10} NO QUANTIZATION INFO {tens.name}")
                 else:
                     print(
-                        "    Output %02d %10s min=%s max=%s scale=%s zero_point=%s %s"
-                        % (idx, tens.dtype, q.min, q.max, q.scale_f32, q.zero_point, tens.name)
+                        f"    Output {idx:02d} {tens.dtype!s:>10}"
+                        f" min={q.min} max={q.max} scale={q.scale_f32!s} zero_point={q.zero_point} {tens.name}"
                     )
             print()
 
@@ -507,10 +534,13 @@ class Graph:
         self.subgraphs = []
         self.metadata = []
         self.memory_used = {}
-        self.bits_per_element = {}
-        self.total_size = {}
-        self.total_elements = {}
+        self.total_original_weights = 0
+        self.total_npu_weights = 0
+        self.total_npu_encoded_weights = 0
         self.weight_cache = None  # See CompressedWeightCache
+        self.bandwidths = 0
+        self.macs = 0
+        self.cycles = 0
 
     def get_root_subgraph(self):
         return self.subgraphs[0]
@@ -531,9 +561,9 @@ class Graph:
         for sg in self.subgraphs:
             sg.print_operators()
 
-    def print_graph(self):
+    def print_graph(self, label=None):
         for sg in self.subgraphs:
-            sg.print_graph()
+            sg.print_graph(label)
 
     def print_graph_with_tensors(self):
         for sg in self.subgraphs:
