@@ -1,4 +1,4 @@
-# Copyright (C) 2020 Arm Limited or its affiliates. All rights reserved.
+# Copyright (C) 2020-2021 Arm Limited or its affiliates. All rights reserved.
 #
 # SPDX-License-Identifier: Apache-2.0
 #
@@ -21,29 +21,25 @@ import enum
 from .debug_database import DebugDatabase
 from .nn_graph import Pass
 from .nn_graph import PassPlacement
-from .operation import create_avgpool_nop
 from .operation import NpuBlockType
 from .operation import Op
+from .operation_util import create_avgpool_nop
 from .tensor import TensorPurpose
 
 
 class PassFlags(enum.Flag):
     Empty = 0
-    Pre = 1
-    Main = 2
-    Post = 4
-    Mac = 8
-    Dma = 32
-    ElementWise = 256
-    Npu = 512
-    Cpu = 1024
-    StartupInit = 2048
-    MemoryOnly = 4096
-    PostFusingLimited = 8192
+    Main = 1
+    Post = 2
+    Mac = 4
+    ElementWise = 8
+    Npu = 16
+    Cpu = 32
+    StartupInit = 64
+    MemoryOnly = 128
+    PostFusingLimited = 256
 
 
-npu_pre_ops = set((Op.SplitSliceRead,))
-
 mac_main_ops = set(
     (
         # convolutions
@@ -65,14 +61,13 @@ mac_main_ops = set(
         Op.AvgPool,
         Op.MaxPool,
         Op.ReduceSum,
-        # deconvolution
-        Op.ResizeBilinear,
     )
-)
+    # resize ops use pooling operations unless explicitly converted to other operations prior to pass packing
+) | Op.op_set(Op.is_resize_op)
 
 binary_elem_wise_main_ops = Op.op_set(Op.is_binary_elementwise_op)
 
-unary_elem_wise_main_ops = Op.op_set(Op.is_unary_elementwise_op)  # Unary element-wise operations
+unary_elem_wise_main_ops = Op.op_set(Op.is_unary_elementwise_op)
 
 elem_wise_main_ops = binary_elem_wise_main_ops | unary_elem_wise_main_ops
 
@@ -81,7 +76,7 @@ npu_post_ops = activation_ops
 
 npu_post_fuse_limited_ops = set(
     # Set of post operators that should not be fused with main/elementwise ops
-    (Op.ConcatSliceWrite, Op.Sigmoid, Op.Tanh, Op.Quantize)
+    (Op.Sigmoid, Op.Tanh, Op.Quantize)
 )
 
 elem_wise_ops = elem_wise_main_ops | activation_ops | set((Op.Sigmoid, Op.Tanh))
@@ -90,9 +85,15 @@ elem_wise_ops = elem_wise_main_ops | activation_ops | set((Op.Sigmoid, Op.Tanh))
 quantization_ops = set((Op.Dequantize, Op.Max, Op.Min))
 cpu_ops = set((Op.Softmax, Op.LRN, Op.Shape, Op.Pad, Op.AddN)) | quantization_ops
 
-npu_dma_ops = set((Op.DMA,))
 startup_init_ops = set((Op.Const, Op.Placeholder, Op.SubgraphInput))
-memory_only_ops = set((Op.Squeeze, Op.Reshape, Op.QuantizedReshape, Op.ExpandDims,))
+memory_only_ops = set(
+    (
+        Op.Squeeze,
+        Op.Reshape,
+        Op.QuantizedReshape,
+        Op.ExpandDims,
+    )
+)
 
 
 test_sequence = [
@@ -100,7 +101,7 @@ test_sequence = [
         # ops_set
         npu_post_ops,
         # incompatible_pack_flags
-        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.Pre | PassFlags.Main,
+        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.Main,
         # flags_to_set
         PassFlags.Npu | PassFlags.Post,
         # flags_to_clear
@@ -110,7 +111,7 @@ test_sequence = [
         # ops_set
         npu_post_fuse_limited_ops,
         # incompatible_pack_flags
-        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.Pre | PassFlags.Main,
+        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.Main | PassFlags.PostFusingLimited,
         # flags_to_set
         PassFlags.Npu | PassFlags.PostFusingLimited,
         # flags_to_clear
@@ -120,12 +121,7 @@ test_sequence = [
         # ops_set
         mac_main_ops,
         # incompatible_pack_flags
-        PassFlags.Cpu
-        | PassFlags.MemoryOnly
-        | PassFlags.ElementWise
-        | PassFlags.Pre
-        | PassFlags.Main
-        | PassFlags.PostFusingLimited,
+        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.ElementWise | PassFlags.Main | PassFlags.PostFusingLimited,
         # flags_to_set
         PassFlags.Npu | PassFlags.Mac | PassFlags.Main,
         # flags_to_clear
@@ -135,37 +131,12 @@ test_sequence = [
         # ops_set
         elem_wise_main_ops,
         # incompatible_pack_flags
-        PassFlags.Cpu
-        | PassFlags.MemoryOnly
-        | PassFlags.Mac
-        | PassFlags.Pre
-        | PassFlags.Main
-        | PassFlags.PostFusingLimited,
+        PassFlags.Cpu | PassFlags.MemoryOnly | PassFlags.Mac | PassFlags.Main | PassFlags.PostFusingLimited,
         # flags_to_set
         PassFlags.Npu | PassFlags.ElementWise | PassFlags.Main,
         # flags_to_clear
         PassFlags.Empty,
     ),
-    (
-        # ops_set
-        npu_pre_ops,
-        # incompatible_pack_flags
-        PassFlags.Cpu | PassFlags.MemoryOnly,
-        # flags_to_set
-        PassFlags.Npu | PassFlags.Mac | PassFlags.Pre | PassFlags.ElementWise,
-        # flags_to_clear
-        PassFlags.Empty,
-    ),
-    (
-        # ops_set
-        npu_dma_ops,
-        # incompatible_pack_flags
-        PassFlags.Cpu | PassFlags.MemoryOnly,
-        # flags_to_set
-        PassFlags.Npu | PassFlags.Dma,
-        # flags_to_clear
-        PassFlags.Empty,
-    ),
     (
         # ops_set
         startup_init_ops,
@@ -231,9 +202,11 @@ def pack_into_passes(nng, arch, verbose_packing=False):
                 ofm_tensor = op.ofm
                 if ofm_tensor is None:
                     ofm_tensor = op.outputs[0]
-                build_pass((op,), ofm_tensor)
+                ofm_shape = op.ofm_shapes[0] if op.run_on_npu else None
 
-    def build_pass(start_ops_to_process, ofm_tensor=None):
+                build_pass((op,), ofm_tensor, ofm_shape)
+
+    def build_pass(start_ops_to_process, ofm_tensor=None, ofm_shape=None):
         reverse_ops_list = []
         curr_flags = PassFlags.Empty
         npu_block_type = NpuBlockType.Default
@@ -242,6 +215,7 @@ def pack_into_passes(nng, arch, verbose_packing=False):
         input_set = set()
         ifm_tensor = None
         primary_op = None
+        ifm_shapes = None
 
         to_process = collections.deque()
         for start_op in start_ops_to_process:
@@ -277,36 +251,19 @@ def pack_into_passes(nng, arch, verbose_packing=False):
                             ):
                                 assert len(curr_op.inputs) >= 1
                                 ifm_tensor = curr_op.ifm
+                                ifm_shapes = curr_op.ifm_shapes.copy()
                                 assert ifm_tensor is not None, "IFM missing in {}".format(curr_op)
                                 assert ifm_tensor.purpose == TensorPurpose.FeatureMap
 
-                        if flags_to_set & PassFlags.Dma:
-                            # DMAs are special - Output buffers need to be preserved as intermediates,
-                            # if the pass consumes the results
-                            if tens is not None:
-                                reverse_intermediates.append(tens)
-
                         if operation_set is None:
                             print("Warning:", curr_op.type, "operation is unknown or unsupported, placing on CPU")
 
                         for inp in reversed(curr_op.inputs):
                             if inp is None:
                                 continue
-                            can_pack = True
-                            if len(inp.ops) == 1:
-                                next_op = inp.ops[0]
-                                for outp in next_op.outputs:
-                                    consumers = outp.consumers()
-                                    if len(consumers) > 1 or (len(consumers) == 1 and consumers[0] != curr_op):
-                                        can_pack = False
-                                        break
+                            if can_pack(inp, curr_op):
+                                to_process.append((inp.ops[0], inp))
                             else:
-                                can_pack = False
-
-                            if can_pack:
-                                to_process.append((next_op, inp))
-                            else:
-                                assert inp is not None
                                 input_set.add(inp)
 
                         break
@@ -323,7 +280,7 @@ def pack_into_passes(nng, arch, verbose_packing=False):
 
         is_element_wise = True
         for op in reverse_ops_list:
-            if op.type not in elem_wise_ops and op.type not in npu_dma_ops:
+            if op.type not in elem_wise_ops and op.type:
                 is_element_wise = False
                 break
 
@@ -366,11 +323,6 @@ def pack_into_passes(nng, arch, verbose_packing=False):
             for inp in primary_op.inputs:
                 if inp is None:
                     continue
-                if len(inp.ops) == 1 and inp.ops[0].type == Op.DMA and inp.purpose == TensorPurpose.FeatureMap:
-                    src_op = inp.ops[0]
-                    if src_op in input_ops_list:
-                        inp = src_op.inputs[0]
-                        input_ops_list.remove(src_op)
                 add_input_list(inp, input_set, input_refcounts, lut_list, ordered_input_list)
             input_ops_list.remove(primary_op)
 
@@ -380,9 +332,6 @@ def pack_into_passes(nng, arch, verbose_packing=False):
                 add_input_list(inp, input_set, input_refcounts, lut_list, ordered_input_list)
 
         name = ops_list[0].name
-        non_dma_ops = [op for op in ops_list if op.type != Op.DMA]
-        if non_dma_ops:
-            name = non_dma_ops[0].name
         ps = Pass(name, placement, is_element_wise, npu_block_type)
         ps.ops = ops_list
         ps.primary_op = primary_op
@@ -397,11 +346,28 @@ def pack_into_passes(nng, arch, verbose_packing=False):
 
             if len(ps.inputs) > 2:
                 ps.ifm_tensor = ps.inputs[-2]
+
+            # Get the corresponding ifm_shapes
+            for op in input_ops_list + [primary_op]:
+                if op.run_on_npu:
+                    if ps.ifm_tensor == op.ifm:
+                        ps.ifm_shapes.append(op.ifm_shapes[0])
+                    elif ps.ifm_tensor == op.ifm2:
+                        ps.ifm_shapes.append(op.ifm_shapes[1])
+
+                    if ps.ifm2_tensor == op.ifm:
+                        ps.ifm_shapes.append(op.ifm_shapes[0])
+                    elif ps.ifm2_tensor == op.ifm2:
+                        ps.ifm_shapes.append(op.ifm_shapes[1])
         else:
             ps.ifm_tensor = ifm_tensor
             ps.ifm2_tensor = None
+            if ps.primary_op is not None and ps.primary_op.run_on_npu:
+                ps.ifm_shapes.append(ifm_shapes[0])
 
         ps.ofm_tensor = ofm_tensor
+        ps.ofm_shapes.append(ofm_shape)
+
         assert ps.placement != PassPlacement.Npu or ps.ofm_tensor is not None
         ps.weight_tensor = ps.get_primary_op_ifm_weights()[1]
         ps.scale_tensor = ps.get_primary_op_ifm_weights_biases_ofm()[2]
@@ -427,7 +393,7 @@ def pack_into_passes(nng, arch, verbose_packing=False):
                 visit_op(op, tens)
 
     def create_primary_op(op_list):
-        if any(op.type in (npu_pre_ops | npu_post_ops | npu_post_fuse_limited_ops) and op.run_on_npu for op in op_list):
+        if any(op.type in (npu_post_ops | npu_post_fuse_limited_ops) and op.run_on_npu for op in op_list):
             # Configure a 1x1 AvgPool and attach the op onto it
             op = op_list[0]
             inp = op.inputs[0]
@@ -436,6 +402,10 @@ def pack_into_passes(nng, arch, verbose_packing=False):
             avgpool_out = inp.clone("_avgpooled")
             avgpool_out.consumer_list.append(op)
             avgpool_op.set_output_tensor(avgpool_out)
+            avgpool_op.ifm_shapes = op.ifm_shapes.copy()
+            avgpool_op.ofm_shapes = op.ofm_shapes.copy()
+            avgpool_op.read_offsets = op.read_offsets.copy()
+            avgpool_op.read_shapes = op.read_shapes.copy()
 
             op.inputs[0] = avgpool_out
             op_list.insert(0, avgpool_op)
@@ -445,6 +415,27 @@ def pack_into_passes(nng, arch, verbose_packing=False):
 
         return None
 
+    def can_pack(inp, curr_op):
+        if len(inp.ops) == 1:
+            next_op = inp.ops[0]
+            for outp in next_op.outputs:
+                consumers = outp.consumers()
+                if len(consumers) > 1 or (len(consumers) == 1 and consumers[0] != curr_op):
+                    return False
+
+            # There cannot be any reshaping between next_op ofm and corresponding curr_op ifm
+            if len(curr_op.ifm_shapes) != 0 and len(next_op.ofm_shapes) != 0:
+                if inp == curr_op.ifm and next_op.ofm_shapes[0] != curr_op.ifm_shapes[0]:
+                    return False
+                elif (
+                    curr_op.ifm2 is not None and inp == curr_op.ifm2 and next_op.ofm_shapes[0] != curr_op.ifm_shapes[1]
+                ):
+                    return False
+        else:
+            return False
+
+        return True
+
     def add_input_list(inp_to_add, inp_set, inp_refcnts, lut_list, ordered_inp_list):
         if inp_to_add in inp_set:
             if inp_refcnts[inp_to_add] == 0:
@@ -469,7 +460,76 @@ def pack_into_passes(nng, arch, verbose_packing=False):
             startup_ps.outputs = [op.outputs[0] for op in startup_list]  # Need to fixup the outputs
             startup_ps.name = "startup_weight_initialisation"
 
-        sg.passes = list(reversed(reverse_pass_list))
+        # Graphs with both CPU and NPU ops might not have an optimal order in
+        # the pass list due to how the graph is traversed (depth first search).
+        # This can result in more context switching between CPU and NPU.
+        # Try to optmize this by moving/grouping CPU ops where that is possible.
+        # Criteria for CPU pass to be moved:
+        #
+        # 1) CPU passes that only depends on sg.input_tensor can be
+        #    moved to the top of the list.
+        #
+        # 2) A CPU pass X is allowed to be grouped together with CPU pass Y
+        #    if there is no NPU pass between pass X and pass Y that depends
+        #    on output from pass X or a MemoryOnly pass.
+        #
+        # Criteria 2 will try to move as many CPU passes towards the bottom of
+        # the list.
+
+        pass_list_top = []
+        pass_list = []
+
+        # Filter out early passes from the rest
+        for ps in list(reversed(reverse_pass_list)):
+            if startup_ps == ps:
+                # startup pass belongs in the top
+                pass_list_top.insert(0, ps)
+                continue
+
+            if (
+                ps.placement == PassPlacement.Cpu
+                and ps.ops[0].ifm in sg.input_tensors
+                and (ps.ops[0].ifm2 in sg.input_tensors or ps.ops[0].ifm2 is None)
+            ):
+                # This CPU pass only depends on sg.input_tensors
+                pass_list_top.append(ps)
+            else:
+                # Add pass to the list that will be sorted in the next step
+                pass_list.append(ps)
+
+        # Sort the rest of the list based on critera 2.
+        # Search from bottom of list and when a CPU pass is found
+        # search forward in the list and see if it is possible to join another CPU pass.
+        last_idx = len(pass_list) - 1
+        for cpu_ps in reversed(pass_list):
+            if cpu_ps.placement != PassPlacement.Cpu:
+                continue
+            # CPU pass found, search forward and move pass if possible
+            idx = pass_list.index(cpu_ps)
+            for next_ps in pass_list[idx + 1 :]:
+                if next_ps.placement == PassPlacement.Cpu:
+                    # It is possible to move the CPU pass
+                    pass_list.remove(cpu_ps)
+                    insert_index = pass_list.index(next_ps)
+                    pass_list.insert(insert_index, cpu_ps)
+                    break
+
+                if (
+                    cpu_ps.ops[0].ofm in [next_ps.ops[0].ifm, next_ps.ops[0].ifm2]
+                    or next_ps.placement == PassPlacement.MemoryOnly
+                ):
+                    # Not possible to move
+                    break
+
+                if pass_list.index(next_ps) == last_idx:
+                    # Last element, ok to move the CPU pass
+                    pass_list.remove(cpu_ps)
+                    pass_list.append(cpu_ps)
+                    break
+
+        pass_list_top.extend(pass_list)
+
+        sg.passes = pass_list_top
         sg.build_pass_links()
 
     if verbose_packing:
