@@ -1,4 +1,4 @@
-# Copyright (C) 2020 Arm Limited or its affiliates. All rights reserved.
+# Copyright (C) 2020-2021 Arm Limited or its affiliates. All rights reserved.
 #
 # SPDX-License-Identifier: Apache-2.0
 #
@@ -21,9 +21,12 @@ import numpy as np
 from flatbuffers import encode
 from flatbuffers.builder import UOffsetTFlags
 
+from .errors import VelaError
 from .nn_graph import PassPlacement
 from .operation import Op
+from .reader_util import align_inputs_indices
 from .tensor import MemType
+from .tensor import TensorPurpose
 from .tflite import Buffer
 from .tflite import Metadata
 from .tflite import Model
@@ -36,7 +39,7 @@ from .tflite_mapping import builtin_operator_inv_map
 from .tflite_mapping import BuiltinOperator
 from .tflite_mapping import datatype_inv_map
 
-# ugh, the python flatbuffer interface is missing a method to add in file identifier. patching it in here:
+# the python flatbuffer interface is missing a method to add in file identifier. patching it in here:
 
 tflite_version = 3
 tflite_file_identifier = "TFL" + str(tflite_version)
@@ -44,7 +47,7 @@ tflite_file_identifier = "TFL" + str(tflite_version)
 
 def FinishWithFileIdentifier(self, rootTable, fid):
     if fid is None or len(fid) != 4:
-        raise Exception("fid must be 4 chars")
+        raise VelaError("FileIdentifier must be 4 chars")
 
     flags = N.Uint8Flags
     prepSize = 4
@@ -76,7 +79,7 @@ class TFLiteSerialiser:
         self.scratch_fast_buf_id = 1  # Always assign scratch_fast to buffer 1
         self.buffers_to_write = []  # have an empty array there
 
-        self.ops_to_ignore = set((Op.Const, Op.Placeholder, Op.SubgraphInput))
+        self.ops_to_ignore = (Op.Const, Op.Placeholder, Op.SubgraphInput)
 
         self.tensors_to_reshape = {}
 
@@ -87,6 +90,8 @@ class TFLiteSerialiser:
             for ps in sg.passes:
                 for op in ps.ops:
                     if op.type not in self.ops_to_ignore:
+                        # swap from nng input indexing to TensorFlow Lite input indexing
+                        self.align_nng_inputs_to_tflite(op)
                         all_ops.append(op)
                     if op.type.is_conv2d_op() or op.type.is_depthwise_conv2d_op():
                         # If values are None op has non-constant weights
@@ -101,6 +106,11 @@ class TFLiteSerialiser:
         self.operator_codes = sorted(set((op.type, op.attrs.get("custom_code", "")) for op in all_ops))
         self.operator_code_map = {}
 
+    def align_nng_inputs_to_tflite(self, op):
+        from_indices = op.type.info.indices
+        _, _, to_indices = builtin_operator_inv_map[op.type]
+        op.inputs = align_inputs_indices(from_indices, to_indices, op.inputs)
+
     def write_byte_vector(self, v, alignment=1):
         builder = self.builder
         builder.StartVector(1, len(v), alignment)
@@ -167,23 +177,32 @@ class TFLiteSerialiser:
         builder = self.builder
         custom_code_offset = None
         if op_type == Op.Custom:
-            tf_code, opt_serializer = builtin_operator_inv_map[op_type]
+            tf_code, opt_serializer, _ = builtin_operator_inv_map[op_type]
             custom_code_offset = builder.CreateString(custom_code)
         else:
             assert (
                 op_type in builtin_operator_inv_map
             ), "Vela does not contain a mapping to serialise {} operator to a TensorFlow Lite operator".format(op_type)
-            tf_code, opt_serializer = builtin_operator_inv_map[op_type]
+            tf_code, opt_serializer, _ = builtin_operator_inv_map[op_type]
 
-            if tf_code == BuiltinOperator.CUSTOM:
+            if op_type == Op.CustomNpuOp:
                 assert (
-                    op_type == Op.CustomNpuOp
+                    tf_code == BuiltinOperator.CUSTOM
                 ), "Vela only supports serialising NpuOp operators as TensorFlow Lite Custom operators"
                 custom_code_offset = builder.CreateString("ethos-u")
 
-        self.operator_code_map[op_type] = (idx, tf_code, opt_serializer)
+        # there can be multiple different types of 3rd party custom operators (i.e. non-"ethos-u" ones). therefore we
+        # need to add an extra level of indirection to this particular entry in the operator_code_map to allow for the
+        # correct lookup later on
+        if op_type == Op.Custom:
+            if op_type not in self.operator_code_map:
+                self.operator_code_map[op_type] = {}
+            self.operator_code_map[op_type][custom_code] = (idx, tf_code, opt_serializer)
+        else:
+            self.operator_code_map[op_type] = (idx, tf_code, opt_serializer)
 
         OperatorCode.OperatorCodeStart(builder)
+        OperatorCode.OperatorCodeAddDeprecatedBuiltinCode(builder, tf_code if tf_code < 127 else 127)
         OperatorCode.OperatorCodeAddBuiltinCode(builder, tf_code)
         if custom_code_offset is not None:
             OperatorCode.OperatorCodeAddCustomCode(builder, custom_code_offset)
@@ -193,6 +212,7 @@ class TFLiteSerialiser:
     def serialise_quantization_parameters(self, quant):
         builder = self.builder
 
+        qp = None
         min = None
         max = None
         scale = None
@@ -207,23 +227,25 @@ class TFLiteSerialiser:
             if quant.zero_point is not None:
                 zero_point = self.write_long_vector(make_vector(quant.zero_point))
 
-        QuantizationParameters.QuantizationParametersStart(builder)
-        if min is not None:
-            QuantizationParameters.QuantizationParametersAddMin(builder, min)
-        if max is not None:
-            QuantizationParameters.QuantizationParametersAddMax(builder, max)
-        if scale is not None:
-            QuantizationParameters.QuantizationParametersAddScale(builder, scale)
-        if zero_point is not None:
-            QuantizationParameters.QuantizationParametersAddZeroPoint(builder, zero_point)
-        return QuantizationParameters.QuantizationParametersEnd(builder)
+            QuantizationParameters.QuantizationParametersStart(builder)
+            if min is not None:
+                QuantizationParameters.QuantizationParametersAddMin(builder, min)
+            if max is not None:
+                QuantizationParameters.QuantizationParametersAddMax(builder, max)
+            if scale is not None:
+                QuantizationParameters.QuantizationParametersAddScale(builder, scale)
+            if zero_point is not None:
+                QuantizationParameters.QuantizationParametersAddZeroPoint(builder, zero_point)
+            if quant.quant_dim is not None:
+                QuantizationParameters.QuantizationParametersAddQuantizedDimension(builder, quant.quant_dim)
+            qp = QuantizationParameters.QuantizationParametersEnd(builder)
+
+        return qp
 
     def serialise_tensor(self, tens):
         builder = self.builder
         tens_shape = tens.shape
-        values = tens.quant_values
-        if values is None:
-            values = tens.values
+        values = tens.values
 
         if values is None:
             values = np.empty(shape=(0), dtype=np.uint8)
@@ -248,7 +270,9 @@ class TFLiteSerialiser:
         # Empty buffers should be kept unique for TensorFlow Lite Micro
         Tensor.TensorAddBuffer(builder, buf_id)
         Tensor.TensorAddName(builder, name)
-        Tensor.TensorAddQuantization(builder, quant)
+        if quant is not None:
+            Tensor.TensorAddQuantization(builder, quant)
+        Tensor.TensorAddIsVariable(builder, tens.is_variable)
 
         res = Tensor.TensorEnd(builder)
         return res
@@ -256,12 +280,20 @@ class TFLiteSerialiser:
     def serialise_operator(self, op):
         builder = self.builder
 
-        inputs_offset = self.write_int_vector([self.tensor_map[tens] for tens in op.inputs if tens in self.tensor_map])
+        inputs_offset = self.write_int_vector(
+            [self.tensor_map[tens] if tens in self.tensor_map else -1 for tens in op.inputs]
+        )
         outputs_offset = self.write_int_vector(
             [self.tensor_map[tens] for tens in op.outputs if tens in self.tensor_map]
         )
+        intermediates_offset = self.write_int_vector(
+            [self.tensor_map[tens] for tens in op.intermediates if tens in self.tensor_map]
+        )
 
-        op_idx, tflop, opt_serializer = self.operator_code_map[op.type]
+        if op.type == Op.Custom:
+            op_idx, tflop, opt_serializer = self.operator_code_map[op.type][op.attrs.get("custom_code", "")]
+        else:
+            op_idx, tflop, opt_serializer = self.operator_code_map[op.type]
 
         builtin_opt_offset = None
         custom_opt_offset = None
@@ -287,6 +319,7 @@ class TFLiteSerialiser:
         Operator.OperatorAddOpcodeIndex(builder, op_idx)
         Operator.OperatorAddInputs(builder, inputs_offset)
         Operator.OperatorAddOutputs(builder, outputs_offset)
+        Operator.OperatorAddIntermediates(builder, intermediates_offset)
 
         if builtin_opt_offset is not None:
             Operator.OperatorAddBuiltinOptionsType(builder, opt_serializer.builtin_opt_type)
@@ -315,13 +348,13 @@ class TFLiteSerialiser:
         # This allows us to serialise tensors which arent attached to any specific ops,
         # e.g. due to an empty graph containing no ops
         for op in all_ops + placeholder_ops:
-            for tens in op.inputs + op.outputs:
+            for tens in op.inputs + op.outputs + op.intermediates:
                 if tens is not None:
                     tensor_set.add(tens)
 
         all_tensors = [tens for nm, idx, tens in sorted((tens.name, idx, tens) for idx, tens in enumerate(tensor_set))]
 
-        scratch_tensors = [tens for tens in all_tensors if tens.name.endswith("scratch")]
+        scratch_tensors = [tens for tens in all_tensors if tens.purpose is TensorPurpose.Scratch]
 
         if len(scratch_tensors) == 0:
             scratch_tensor = None
@@ -399,16 +432,19 @@ class TFLiteSerialiser:
         subgraph_idx = np.int32(len(self.subgraphs_to_write))  # Only 1 supported currently
         nbr_tensors = np.int32(len(self.tensor_map))
 
-        # An offset of -1 indicates that the tensor will be allocated online by Tensorflow Lite Micro
-        offsets = [np.int32(-1)] * nbr_tensors
+        if not any([name == b"OfflineMemoryAllocation" for name, _ in self.nng.metadata]):
+            # An offset of -1 indicates that the tensor will be allocated online by Tensorflow Lite Micro
+            offsets = [np.int32(-1)] * nbr_tensors
 
-        # Ensure that the order of the offsets match the order of the tensors
-        for tens, idx in self.tensor_map.items():
-            # Set offsets for tensor allocated in Tensor Arena or in the scratch_fast area
-            if tens.mem_type in set((MemType.Scratch, MemType.Scratch_fast)):
-                offsets[idx] = np.int32(tens.address) if tens.address is not None else np.int32(0)
+            # Ensure that the order of the offsets match the order of the tensors
+            for tens, idx in self.tensor_map.items():
+                # Set offsets for tensor allocated in Tensor Arena or in the scratch_fast area
+                if tens.mem_type in (MemType.Scratch, MemType.Scratch_fast):
+                    offsets[idx] = np.int32(tens.address) if tens.address is not None else np.int32(0)
 
-        self.nng.metadata.append(("OfflineMemoryAllocation", np.array([version, subgraph_idx, nbr_tensors] + offsets)))
+            self.nng.metadata.append(
+                ("OfflineMemoryAllocation", np.array([version, subgraph_idx, nbr_tensors] + offsets))
+            )
 
         metadata_list = []
         for name, buffer in self.nng.metadata:
